"""Test repository with multiple dvc.yaml files."""

import os
import subprocess

import git
import pytest
import yaml
from typer.testing import CliRunner

from paraffin.cli import app

runner = CliRunner()

dvc_yaml_dict = {
    "stages": {
        "01_preprocessing": {
            "outs": ["output"],
            "cmd": [
                """bash -euo pipefail << EOF

# add bash code here
mkdir -p output && echo "Hello World" > output/hello.txt

EOF"""
            ],
        }
    }
}


@pytest.fixture
def nested_project(tmp_path):
    subprocess.check_call(["git", "init"], cwd=tmp_path)
    subprocess.check_call(["dvc", "init"], cwd=tmp_path)

    (tmp_path / "dvc.yaml").touch()
    exp1 = tmp_path / "exp1"
    exp1.mkdir()

    (exp1 / "dvc.yaml").write_text(yaml.dump(dvc_yaml_dict))

    repo = git.Repo(tmp_path)
    repo.git.add(all=True)
    repo.index.commit("Initial commit")

    return tmp_path


def test_053(nested_project):
    os.chdir(nested_project)
    # subprocess.check_call(["dvc repro"], shell=True)

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0

    result = runner.invoke(app, "worker")
    assert result.exit_code == 0

    assert (
        nested_project / "exp1" / "output" / "hello.txt"
    ).read_text() == "Hello World\n"



"""Test an issue with full path instead of file dependencies."""

import os
import subprocess

import git
import pytest
from typer.testing import CliRunner

from paraffin.cli import app

runner = CliRunner()


CONTENT = """
import zntrack
import pathlib
import zntrack.examples
import random
import subprocess


class ReadFile(zntrack.Node):
    path: pathlib.Path = zntrack.deps_path()
    data: float = zntrack.outs()

    def run(self):
        with open(self.path / "data.csv") as f:
            self.data = sum(map(float, f.read().split(",")))


if __name__ == "__main__":
    data = pathlib.Path("data")
    data.mkdir(exist_ok=True)
    data_file1 = pathlib.Path("data/data.csv")
    data_file1.unlink(missing_ok=True)
    with data_file1.open("w") as f:
        f.write(",".join(str(random.randint(1, 10000)) for _ in range(5)))

    subprocess.check_call(["dvc", "add", "data"])

    proj = zntrack.Project()
    with proj:
        a = ReadFile(path=pathlib.Path("data"))

    proj.build()
"""


@pytest.fixture
def repo(tmp_path):
    os.chdir(tmp_path)
    repo = git.Repo.init(tmp_path)
    # run dvc init in the repo
    subprocess.run(["dvc", "init"], cwd=tmp_path)

    (tmp_path / "main.py").write_text(CONTENT)

    subprocess.check_call(["python", "main.py"])
    # make a git commit

    repo.index.add(["main.py"])
    repo.index.commit("Initial commit")

    return repo


def test_053(repo):
    # subprocess.check_call(["dvc repro"], shell=True)

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0

    result = runner.invoke(app, "worker")
    assert result.exit_code == 0



import datetime
import logging
import os
import socket
import threading
import time
import typing as t
import webbrowser

import git
import typer
import uvicorn

from paraffin.db import (
    close_worker,
    complete_job,
    find_cached_job,
    get_job,
    register_worker,
    save_graph_to_db,
    update_worker,
)
from paraffin.stage import checkout, get_lock, repro
from paraffin.ui.app import app as webapp
from paraffin.utils import (
    detect_zntrack,
    get_custom_queue,
    get_stage_graph,
    update_gitignore,
)

log = logging.getLogger(__name__)

app = typer.Typer()


def spawn_worker(
    name: str, queues, experiment: str, job: str, timeout: float, db: str, workers: dict
):
    worker_id = register_worker(
        name=name,
        machine=socket.gethostname(),
        db_url=db,
        cwd=os.getcwd(),
        pid=os.getpid(),
    )
    workers[worker_id] = None
    log.info(f"Listening on queues: {queues}")

    last_seen = datetime.datetime.now()
    try:
        while True:
            job_obj = get_job(
                db_url=db,
                queues=queues,
                worker_id=worker_id,
                experiment=experiment,
                job_name=job,
            )

            if job_obj is None:
                remaining_seconds = (
                    timeout - (datetime.datetime.now() - last_seen).seconds
                )
                if remaining_seconds <= 0:
                    log.info("Timeout reached - exiting.")
                    break
                time.sleep(1)
                log.info(
                    "No more job found"
                    f" - sleeping until closing in {remaining_seconds} seconds"
                )
                continue
            last_seen = datetime.datetime.now()

            update_worker(worker_id, status="running", db_url=db)
            workers[worker_id] = job_obj["id"]

            # This will search the DB and not rely on DVC run cache to determine if
            #  the job is cached so this can easily work across directories
            cached_job = False
            if job_obj["cache"] and detect_zntrack(job_obj) and not job_obj["force"]:
                stage_lock, deps_hash = get_lock(job_obj["name"])
                cached_job = find_cached_job(deps_cache=deps_hash, db_url=db)
            if cached_job:
                log.info(
                    f"Job '{job_obj['name']}' is cached and dvc.lock is available."
                )
                returncode, stdout, stderr = checkout(
                    stage_lock, cached_job["lock"], job_obj["name"]
                )
                if returncode == 404:
                    # TODO: we need to ensure that all deps nodes are checked out!
                    #  this will be important when clone / push.
                    # TODO: this can be the cause for a lock issue!
                    log.warning(
                        "Unable to checkout GIT tracked files"
                        f" for job '{job_obj['name']}'"
                    )
                    log.info(f"Running job '{job_obj['name']}'")
                    returncode, stdout, stderr = repro(
                        job_obj["name"], force=job_obj["force"]
                    )  # TODO: this is not tested in CI,
                    #  because it did not raise an error
            else:
                log.info(f"Running job '{job_obj['name']}'")
                # TODO: we need to ensure that all deps nodes are checked out!
                #  this will be important when clone / push.
                # TODO: this can be the cause for a lock issue!
                returncode, stdout, stderr = repro(
                    job_obj["name"], force=job_obj["force"]
                )
            if returncode != 0:
                complete_job(
                    job_obj["id"],
                    status="failed",
                    lock={},
                    stdout=stdout,
                    stderr=stderr,
                    db_url=db,
                )
            else:
                stage_lock, _ = get_lock(job_obj["name"])
                complete_job(
                    job_obj["id"],
                    status="completed",
                    lock=stage_lock,
                    stdout=stdout,
                    stderr=stderr,
                    db_url=db,
                )
            job_obj = None
            update_worker(worker_id, status="idle", db_url=db)

    finally:
        if job_obj is not None:
            complete_job(
                job_obj["id"],
                status="failed",
                lock={},
                stdout="",
                stderr="Worker exited.",
                db_url=db,
            )
        close_worker(id=worker_id, db_url=db)
        workers.pop(worker_id)


@app.command()
def ui(
    port: int = 8000,
    db: str = typer.Option(
        "sqlite:///paraffin.db", help="Database URL.", envvar="PARAFFIN_DB"
    ),
    all: bool = typer.Option(
        False, help="Show all experiments and not just from the current commit."
    ),
):
    """Start the Paraffin web UI."""
    if not all:
        try:
            repo = git.Repo(search_parent_directories=True)
            commit = repo.head.commit
            os.environ["PARAFFIN_COMMIT"] = commit.hexsha
        except git.InvalidGitRepositoryError:
            log.warning(
                "Unable to determine the current commit. Showing all experiments."
            )

    webbrowser.open(f"http://localhost:{port}")
    os.environ["PARAFFIN_DB"] = db
    uvicorn.run(webapp, host="0.0.0.0", port=port)


@app.command()
def worker(
    queues: str = typer.Option(
        "default",
        "--queues",
        "-q",
        envvar="PARAFFIN_QUEUES",
        help="Comma separated list of queues to listen on.",
    ),
    name: str = typer.Option(
        "default", "--name", "-n", help="Specify a custom name for this worker."
    ),
    job: str | None = typer.Option(None, "--job", "-j", help="Job ID to run."),
    experiment: str | None = typer.Option(
        None, "--experiment", "-e", help="Experiment ID to run."
    ),
    timeout: int = typer.Option(
        0,
        "--timeout",
        "-t",
        help="Timeout in seconds before exiting"
        " the worker if no more jobs are in the queue.",
    ),
    db: str = typer.Option(
        "sqlite:///paraffin.db", help="Database URL.", envvar="PARAFFIN_DB"
    ),
    jobs: int = typer.Option(1, "--jobs", "-j", help="Number of jobs to run."),
    delay_between_workers: float = typer.Option(
        0.1, help="Delay between starting workers.", hidden=True
    ),
):
    """Start a paraffin worker to process the queued DVC stages."""
    queues = queues.split(",")
    logging.basicConfig(level=logging.INFO)
    threads = []

    workers = {}
    try:
        for i in range(jobs):
            t = threading.Thread(
                target=spawn_worker,
                args=(name, queues, experiment, job, timeout, db, workers),
                daemon=True,
            )
            threads.append(t)
            t.start()
            time.sleep(delay_between_workers)

        for t in threads:
            t.join()
    finally:
        for worker, job_id in workers.items():
            if job_id is not None:
                complete_job(
                    job_id=job_id,
                    status="failed",
                    lock={},
                    stdout="",
                    stderr="Worker exited.",
                    db_url=db,
                )
            close_worker(id=worker, db_url=db)


@app.command()
def submit(
    names: t.Optional[list[str]] = typer.Argument(
        None, help="Stage names to run. If not specified, run all stages."
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Verbose output."),
    cache: bool = typer.Option(
        False,
        help="Use the paraffin cache in addition to the DVC cache"
        " to checkout cached jobs.",
    ),
    db: str = typer.Option(
        "sqlite:///paraffin.db", help="Database URL.", envvar="PARAFFIN_DB"
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="reproduce pipelines, regenerating its results, even if no changes"
        " were found. See https://dvc.org/doc/command-reference/repro#-f"
        " for more information.",
    ),
    single_item: bool = typer.Option(
        False,
        "--single-item",
        "-s",
        help="reproduce only a single stage by turning off the recursive search for"
        " changed dependencies. See https://dvc.org/doc/command-reference/repro#-s"
        " for more information.",
    ),
):
    """Run DVC stages in parallel."""
    if verbose:
        logging.basicConfig(level=logging.DEBUG)

    if single_item and names is None:
        typer.echo("Cannot use single item without specifying names")
        raise typer.Exit(1)

    # check if the repo has a commit
    repo = git.Repo(search_parent_directories=True)
    if not repo.head.is_valid():
        log.error(
            "Unable to create experiment inside a GIT repository without commits."
        )
        return
    else:
        commit = repo.head.commit
        try:
            origin = repo.remotes.origin.url
        except AttributeError:
            origin = "local"
            log.debug(f"Creating new experiment based on commit '{commit}'")

    log.debug("Getting stage graph")
    graph = get_stage_graph(names=names, force=force, single_item=single_item)

    custom_queues = get_custom_queue()
    update_gitignore(line="paraffin.db")
    save_graph_to_db(
        graph,
        queues=custom_queues,
        commit=commit.hexsha,
        origin=origin,
        machine=socket.gethostname(),
        cache=cache,
        db_url=db,
    )



import os
import pathlib
import shutil
import subprocess

import dvc.cli
import git
import pytest
import zntrack.examples


@pytest.fixture
def check_finished():
    def func(names: list[str] | None = None, exclusive: bool = False) -> bool:
        if exclusive:
            raise NotImplementedError
        cmd = ["dvc", "status"]
        for name in names or []:
            cmd.append(name)
        result = subprocess.run(cmd, capture_output=True, check=True)
        finished = (
            result.stdout.decode().strip() == "Data and pipelines are up to date."
        )
        if not finished:
            print(result.stdout.decode())
        return finished

    return func


@pytest.fixture
def proj_path(tmp_path, request) -> pathlib.Path:
    """temporary directory for testing DVC calls

    Parameters
    ----------
    tmp_path
    request: https://docs.pytest.org/en/6.2.x/reference.html#std-fixture-request

    Returns
    -------
    path to temporary directory

    """
    shutil.copy(request.module.__file__, tmp_path)
    os.chdir(tmp_path)
    git.Repo.init()
    dvc.cli.main(["init"])
    git.Repo().index.commit("Initial commit")

    return tmp_path


@pytest.fixture
def proj01(proj_path) -> zntrack.Project:
    """Project with two independent groups of nodes for testing.

    ```mermaid
    flowchart TD
        node0["data/data.csv"]
        node1["A_SumNodeAttributes"]
        node2["A_X_AddNodeNumbers"]
        node3["A_X_ParamsToOuts"]
        node4["A_X_ParamsToOuts_1"]
        node5["A_Y_AddNodeNumbers"]
        node6["A_Y_ParamsToOuts"]
        node7["A_Y_ParamsToOuts_1"]
        node2-->node1
        node3-->node2
        node4-->node2
        node5-->node1
        node6-->node5
        node7-->node5
        node8["B_SumNodeAttributes"]
        node9["B_X_AddNodeNumbers"]
        node10["B_X_ParamsToOuts"]
        node11["B_X_ParamsToOuts_1"]
        node12["B_Y_AddNodeNumbers"]
        node13["B_Y_ParamsToOuts"]
        node14["B_Y_ParamsToOuts_1"]
        node9-->node8
        node10-->node9
        node11-->node9
        node12-->node8
        node13-->node12
        node14-->node12
    ```
    """
    PARAM = 1  # noqa N806

    with zntrack.Project() as proj:
        for x in ["A", "B"]:
            results = []
            for y in ["X", "Y"]:
                with proj.group(x, y):
                    n1 = zntrack.examples.ParamsToOuts(params=PARAM)
                    n2 = zntrack.examples.ParamsToOuts(params=PARAM)
                    res = zntrack.examples.AddNodeNumbers(numbers=[n1, n2])
                    results.append(res.sum)
            with proj.group(x):
                zntrack.examples.SumNodeAttributes(inputs=results, shift=0)

    proj.build()

    assert len(proj) == 14

    return proj


def proj02(proj_path) -> zntrack.Project:
    PARAM = 1  # noqa N806
    project = zntrack.Project()

    with project.group("A"):
        n = zntrack.examples.ParamsToOuts(params=PARAM)
    for idx in range(3):
        with project.group("A", str(idx)):
            _ = zntrack.examples.AddNodeNumbers(numbers=[n])

    project.build()

    assert len(project) == 4

    return project



import zntrack.examples
from typer.testing import CliRunner

from paraffin.cli import app

runner = CliRunner()


def test_run_job(proj_path, check_finished):
    project = zntrack.Project()

    with project:
        a = zntrack.examples.ParamsToOuts(params=1)
        b = zntrack.examples.ParamsToOuts(params=2)
        c = zntrack.examples.AddNodeNumbers(numbers=[a, b])
        d = zntrack.examples.AddNodeAttributes(a=c.sum, b=c.sum)

        # independent node
        e = zntrack.examples.ParamsToOuts(params=3)

    project.build()

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0

    # run a node without predecessors
    result = runner.invoke(app, f"worker --job {a.name} --experiment 1")
    assert result.exit_code == 0
    assert check_finished([a.name])

    #

    # run a node with predecessors that have not been run yet
    result = runner.invoke(app, f"worker --job {d.name} --experiment 1")
    assert result.exit_code == 0
    assert check_finished([d.name])
    assert not check_finished([e.name])


def test_run_job_submit_twice(proj_path, check_finished):
    project = zntrack.Project()

    with project:
        a = zntrack.examples.ParamsToOuts(params=1)
        b = zntrack.examples.ParamsToOuts(params=2)
        c = zntrack.examples.AddNodeNumbers(numbers=[a, b])
        d = zntrack.examples.AddNodeAttributes(a=c.sum, b=c.sum)

        # independent node
        e = zntrack.examples.ParamsToOuts(params=3)

    project.build()

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    result = runner.invoke(app, "submit")
    assert result.exit_code == 0

    result = runner.invoke(app, f"worker --job {d.name} --experiment 1")
    assert result.exit_code == 0
    assert check_finished([d.name])
    assert not check_finished([e.name])



from paraffin.utils import get_group, replace_node_working_dir


def test_get_group():
    assert get_group("Node") == ([], "Node")
    assert get_group("Node_1") == ([], "Node_1")
    assert get_group("grp_Node") == (["grp"], "Node")
    assert get_group("grp_Node_1") == (["grp"], "Node_1")
    assert get_group("grp_a_Node") == (["grp", "a"], "Node")
    assert get_group("grp_a_Node_1") == (["grp", "a"], "Node_1")


def test_replace_node_working_dir():
    # 1
    inp_nwd = "nodes/grp/MyNode"
    ref_nwd = "nodes/MyNode_1"
    ref_path = "nodes/MyNode_1/node-meta.json"

    assert (
        replace_node_working_dir(ref_path, ref_nwd, inp_nwd).as_posix()
        == "nodes/grp/MyNode/node-meta.json"
    )

    # 2
    inp_nwd = "nodes/MyNode"
    ref_nwd = "nodes/MyNode"
    ref_path = "nodes/MyNode/node-meta.json"

    assert (
        replace_node_working_dir(ref_path, ref_nwd, inp_nwd).as_posix()
        == "nodes/MyNode/node-meta.json"
    )

    # 3
    inp_nwd = "nodes/MyNode_1"
    ref_nwd = "nodes/MyNode"
    ref_path = "nodes/MyNode/node-meta.json"

    assert (
        replace_node_working_dir(ref_path, ref_nwd, inp_nwd).as_posix()
        == "nodes/MyNode_1/node-meta.json"
    )

    # 4
    inp_nwd = "grp/MyNode"
    ref_nwd = "MyNode_1"
    ref_path = "nodes/MyNode_1/node-meta.json"

    assert (
        replace_node_working_dir(ref_path, ref_nwd, inp_nwd).as_posix()
        == "nodes/grp/MyNode/node-meta.json"
    )



import pathlib
import subprocess

import dvc.cli
import pytest
import zntrack
import zntrack.examples
from typer.testing import CliRunner

from paraffin.cli import app

runner = CliRunner()


class ReadFile(zntrack.Node):
    path: pathlib.Path = zntrack.deps_path()
    data: float = zntrack.outs()

    def run(self):
        with open(self.path) as f:
            self.data = sum(map(float, f.read().split(",")))


@pytest.fixture
def proj02(proj_path) -> zntrack.Project:
    """Project with two independent groups of nodes for testing."""

    with zntrack.Project() as proj:
        data_path = pathlib.Path("data")
        data_path.mkdir()
        data_file = data_path / "data.csv"
        data_file.write_text("1,2,3")
        a_1 = ReadFile(path=data_file, name="a_1")
        b_1 = ReadFile(path=data_file, name="b_1")

        _ = zntrack.examples.AddNodeAttributes(a=a_1.data, b=a_1.data, name="a_2")
        _ = zntrack.examples.AddNodeAttributes(a=b_1.data, b=b_1.data, name="b_2")

    proj.build()
    dvc.cli.main(["add", "data/data.csv"])

    return proj


def test_check_finished(proj01, check_finished):
    subprocess.check_call(["dvc", "repro", "A_X_ParamsToOuts"])
    assert check_finished(["A_X_ParamsToOuts"])
    assert not check_finished(["A_Y_ParamsToOuts"])
    subprocess.check_call(["dvc", "repro", "A_Y_ParamsToOuts"])
    assert check_finished(["A_Y_ParamsToOuts"])
    assert not check_finished()


def test_run_all(proj01, caplog, check_finished):
    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0

    assert check_finished()


def test_run_all_multi_jobs(proj01, caplog, check_finished):
    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker", "--jobs", "2"])
    assert result.exit_code == 0

    assert check_finished()


def test_run_selection(proj01, caplog, check_finished):
    result = runner.invoke(app, ["submit", "A_X_ParamsToOuts"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0
    # assert "Running 1 stages" in caplog.text
    # caplog.clear()

    assert check_finished(["A_X_ParamsToOuts"])
    assert not check_finished(["A_Y_ParamsToOuts"])

    result = runner.invoke(app, ["submit", "A_Y_ParamsToOuts"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0
    # # assert "Running 1 stages" in caplog.text
    # # caplog.clear()

    assert check_finished(["A_Y_ParamsToOuts"])
    assert not check_finished()

    result = runner.invoke(app, ["submit", "B_X_AddNodeNumbers", "B_Y_AddNodeNumbers"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0
    # # assert "Running 6 stages" in caplog.text
    # # caplog.clear()

    assert not check_finished()
    assert check_finished(
        [
            "B_X_AddNodeNumbers",
            "B_Y_AddNodeNumbers",
            "B_X_ParamsToOuts",
            "B_Y_ParamsToOuts",
            "B_X_ParamsToOuts_1",
            "B_Y_ParamsToOuts_1",
        ]
    )


def test_run_selection_glob(proj01, caplog, check_finished):
    result = runner.invoke(app, ["submit", "A_X_*"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0

    assert check_finished(
        ["A_X_ParamsToOuts", "A_X_ParamsToOuts_1", "A_X_AddNodeNumbers"]
    )


def test_run_datafile(proj02, caplog, check_finished):
    result = runner.invoke(app, ["submit", "a*"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0
    # assert "Running 2 stages" in caplog.text
    # caplog.clear()
    assert check_finished(["a_1", "a_2"])

    result = runner.invoke(app, ["submit", "b*"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0
    # assert "Running 2 stages" in caplog.text
    # caplog.clear()
    assert check_finished(["b_1", "b_2"])

    assert zntrack.from_rev("a_2").c == 12
    assert zntrack.from_rev("b_2").c == 12

    # modify data file and run to check changed outputs
    data_file = pathlib.Path("data/data.csv")
    data_file.unlink()
    data_file.write_text("4,5,6")

    result = runner.invoke(app, ["submit", "a*"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0

    assert check_finished(["a_1", "a_2"])
    assert not check_finished(["b_1", "b_2"])

    assert zntrack.from_rev("a_2").c == 30
    assert zntrack.from_rev("b_2").c == 12

    result = runner.invoke(app, ["submit", "b*"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0

    assert check_finished(["b_1", "b_2"])
    assert zntrack.from_rev("b_1").data == 15
    assert zntrack.from_rev("b_2").c == 30


def test_run_one_two_many(proj02, check_finished):
    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0

    assert check_finished()



import random

import zntrack
from typer.testing import CliRunner

from paraffin.cli import app

runner = CliRunner()


class RNG(zntrack.Node):
    metrics: dict = zntrack.metrics()

    def run(self):
        self.metrics = {"random": random.random()}


def test_force_rerun(proj_path):
    project = zntrack.Project()

    with project:
        _ = RNG()

    project.build()

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    result = runner.invoke(app, "worker")
    assert result.exit_code == 0

    res_a = RNG.from_rev().metrics["random"]

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    result = runner.invoke(app, "worker")
    assert result.exit_code == 0

    # without force
    assert RNG.from_rev().metrics["random"] == res_a

    result = runner.invoke(app, "submit --force")
    assert result.exit_code == 0
    result = runner.invoke(app, "worker")
    assert result.exit_code == 0

    # with force
    assert RNG.from_rev().metrics["random"] != res_a



import re
from collections import OrderedDict
from pathlib import Path

from paraffin.utils import get_group, replace_node_working_dir


def _extract_node_name(cmd: str) -> str | None:
    """Extract the node name from the command string."""
    node_name_match = re.search(r"--name\s+([\w_]+)", cmd)
    return node_name_match.group(1) if node_name_match else None


def _process_params(params: dict, node_name: str | None) -> dict:
    """Generalize the `params` field by replacing the node name with `<node-name>`."""
    generalized_params = {}
    for file, file_params in params.items():
        generalized_file_params = {}
        for key, value in file_params.items():
            # Only replace keys that match the extracted node name
            if key == node_name:
                generalized_file_params["<node-name>"] = value
            else:
                generalized_file_params[key] = value
        generalized_params[file] = generalized_file_params
    return generalized_params


def _process_deps(deps: list) -> list:
    """Generalize the `deps` field by removing the `path` field."""
    new_deps = []
    for dep in deps:
        new_dep = {"hash": dep["hash"]}
        if "files" in dep:
            new_dep["files"] = dep["files"]  # Keep `files` structure unchanged
        if "md5" in dep:
            new_dep["md5"] = dep["md5"]
        if "hash" in dep:
            new_dep["hash"] = dep["hash"]
        new_deps.append(new_dep)
    return new_deps


def clean_lock(raw: dict) -> dict:
    """Clean the lock file for hashing.

    This function removes the "node-name" from the lock file,
    by cleaning: cmd, deps, outs, and params.
    """
    exp = {k: v for k, v in raw.items() if k in ["cmd", "params", "deps"]}

    # Extract the node name from the `cmd`
    cmd_str = " ".join(raw["cmd"]) if isinstance(raw["cmd"], list) else raw["cmd"]
    node_name = _extract_node_name(cmd_str)

    # Generalize `cmd`: replace the extracted node name with `<node-name>`
    #  This only applies to ZnTrack commands.
    if node_name:
        # TODO: what if the list has more than one element?!
        generalized_cmd = cmd_str.replace(f"--name {node_name}", "--name <node-name>")
        exp["cmd"] = (
            [generalized_cmd] if isinstance(raw["cmd"], list) else generalized_cmd
        )

    # Generalize `params` if present
    if "params" in exp:
        exp["params"] = _process_params(raw["params"], node_name)

    # Generalize `deps` by removing `path`, but keeping `files`
    if "deps" in exp:
        exp["deps"] = _process_deps(raw["deps"])

    return exp


def _ordered_dict_to_dict(od: OrderedDict | dict) -> dict:
    """Convert a nested OrderedDict to a nested dict."""
    return {
        k: _ordered_dict_to_dict(v) if isinstance(v, OrderedDict) else v
        for k, v in od.items()
    }


def transform_lock(inp: dict, ref: dict) -> dict:
    """
    Transform the input lock based on the reference lock to
    produce the correct output lock.

    Parameters
    ----------
    inp : dict
        The input lock containing `cmd`, `params`, and `deps`.
    ref : dict
        The reference lock containing information about executed stages
        and their outputs.

    Returns
    -------
    dict
        The transformed lock that merges information from the input and reference.
    """
    # Extract the node name from `cmd` in the input lock
    inp_node_name_match = re.search(r"--name\s+([\w_]+)", inp["cmd"])
    if inp_node_name_match:
        inp_node_name = inp_node_name_match.group(1)
    else:
        raise ValueError("Node name not found in the input lock.")
    # Extract the node name from `cmd` in the reference lock
    ref_node_name_match = re.search(r"--name\s+([\w_]+)", ref["cmd"])
    if ref_node_name_match:
        ref_node_name = ref_node_name_match.group(1)
    else:
        raise ValueError("Node name not found in the reference lock.")

    # Detect groups
    inp_groups, isolated_inp_node_name = get_group(inp_node_name)
    ref_groups, isolated_ref_node_name = get_group(ref_node_name)

    inp_nwd = Path(*inp_groups, isolated_inp_node_name)
    ref_nwd = Path(*ref_groups, isolated_ref_node_name)

    # Transform the `outs` field using the reference lock
    outs = []
    for ref_out in ref.get("outs", []):
        # update the ref_nwd with the input nwd
        ref_out_path = Path(ref_out["path"])
        updated_out_path = replace_node_working_dir(ref_out_path, ref_nwd, inp_nwd)
        ref_out["path"] = updated_out_path.as_posix()
        outs.append(ref_out)

    inp["outs"] = outs

    return _ordered_dict_to_dict(inp)



import os
import subprocess
from pathlib import Path

from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles

from paraffin.db import (
    db_to_graph,
    get_job_dump,
    get_jobs,
    list_experiments,
    list_workers,
    update_job_status,
)
from paraffin.utils import build_elk_hierarchy

FILE = Path(__file__)

app = FastAPI()

app.mount("/ui", StaticFiles(directory=FILE.parent.parent / "static"), name="ui")
app.mount(
    "/assets",
    StaticFiles(directory=FILE.parent.parent / "static" / "assets"),
    name="assets",
)


@app.get("/")
def read_root():
    return RedirectResponse(url="/ui/index.html")


@app.get("/api/v1/experiments")
def read_experiments():
    db_url = os.environ["PARAFFIN_DB"]
    commit = os.getenv("PARAFFIN_COMMIT")

    return list_experiments(commit=commit, db_url=db_url)


@app.get("/api/v1/graph")
def read_graph(experiment: str):
    db_url = os.environ["PARAFFIN_DB"]

    g = db_to_graph(experiment_id=int(experiment), db_url=db_url)
    return build_elk_hierarchy(g)


@app.get("/api/v1/spawn")
def spawn(
    name: str | None = None, experiment: int | None = None, job: str | None = None
):
    # Build the command
    cmd = ["paraffin", "worker"]
    if name is not None:
        cmd += ["--name", name]
    if experiment is not None:
        cmd += ["--experiment", str(experiment)]
    if job is not None:
        cmd += ["--job", job]

    # open subprocess and forget about it
    subprocess.Popen(cmd)
    return 0

    # works but is slow
    # print(f"Running command: {' '.join(cmd)}")

    # # Start the subprocess and capture stdout and stderr
    # process = subprocess.Popen(
    #     cmd,
    #     stdout=subprocess.PIPE,
    #     stderr=subprocess.PIPE,
    #     text=True,  # Ensure output is decoded as text
    #     bufsize=1,  # Line-buffered output
    #     universal_newlines=True,
    # )

    # # Define a generator to stream the output
    # def generate():
    #     while True:
    #         # Read stdout line by line
    #         output = process.stdout.readline()
    #         if output:
    #             yield f"stdout: {output}"
    #         else:
    #             break

    #         # Read stderr line by line
    #         error = process.stderr.readline()
    #         if error:
    #             yield f"stderr: {error}"
    #         else:
    #             break

    #     # Wait for the process to complete
    #     process.wait()
    #     yield f"Process completed with return code: {process.returncode}"

    # # Return a StreamingResponse to stream the output
    # return StreamingResponse(generate(), media_type="text/plain")


@app.get("/api/v1/job")
def read_job(name: str, experiment: int):
    db_url = os.environ["PARAFFIN_DB"]
    return get_job_dump(job_name=name, experiment_id=int(experiment), db_url=db_url)


# list finished, running, and pending and failed jobs
@app.get("/api/v1/jobs")
def read_jobs(experiment: int):
    db_url = os.environ["PARAFFIN_DB"]
    return get_jobs(experiment_id=int(experiment), db_url=db_url)


@app.get("/api/v1/workers")
def read_workers(id: int | None = None):
    db_url = os.environ["PARAFFIN_DB"]
    return list_workers(db_url=db_url, id=id)


# update_job_status
@app.get("/api/v1/job/update")
def update_job(name: str, experiment: int, status: str, force: bool = False):
    db_url = os.environ["PARAFFIN_DB"]
    print(
        f"Updating job {name} to {status} in experiment {experiment} with force={force}"
    )
    return update_job_status(
        job_name=name,
        experiment_id=int(experiment),
        status=status,
        db_url=db_url,
        force=force,
    )



"""Container for a DVC stage."""

import dataclasses
import json
import logging
import random
import subprocess
import threading
import time
from pathlib import Path

import dvc.api
import yaml
from dvc.lock import LockError
from dvc.stage import PipelineStage
from dvc.stage.cache import _get_cache_hash
from dvc.stage.serialize import to_single_stage_lockfile

from paraffin.lock import clean_lock, transform_lock

log = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True, eq=True)
class PipelineStageDC:
    """Container for a DVC stage."""

    stage: PipelineStage
    status: str
    force: bool

    @property
    def changed(self) -> bool:
        """Check if the stage has changed."""
        return json.loads(self.status) != []

    @property
    def name(self) -> str:
        """Return the name of the stage."""
        return self.stage.addressing

    @property
    def cmd(self) -> str:
        """Return the command of the stage."""
        return self.stage.cmd


def retry(times, exceptions, delay: float = 0, exponential: bool = True):
    """
    Retry Decorator
    Retries the wrapped function/method `times` times if the exceptions listed
    in ``exceptions`` are thrown
    :param times: The number of times to repeat the wrapped function/method
    :type times: Int
    :param Exceptions: Lists of exceptions that trigger a retry attempt
    :type Exceptions: Tuple of Exceptions
    """

    def decorator(func):
        def newfn(*args, **kwargs):
            attempt = 0
            while attempt < times:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    attempt += 1
                    log.warning(f"Caught exception {e} - retrying {attempt}/{times}")
                    sleep_time = delay * (2.0**attempt) if exponential else delay
                    sleep_time *= random.uniform(0, 1.0)
                    time.sleep(sleep_time)
            return func(*args, **kwargs)

        return newfn

    return decorator


@retry(10, (LockError,), delay=0.5)
def get_lock(name: str) -> tuple[dict, str]:
    fs = dvc.api.DVCFileSystem(url=None, rev=None)
    with fs.repo.lock:
        stage = fs.repo.stage.collect(name)[0]
        stage.save(allow_missing=True, run_cache=False)
        stage_lock = to_single_stage_lockfile(stage, with_files=True)
        deps_hash = _get_cache_hash(clean_lock(stage_lock), key=False)

    return stage_lock, deps_hash


def _stream_reader(pipe, callback) -> None:
    """Reads lines from a pipe and calls the callback function."""
    with pipe:
        for line in iter(pipe.readline, ""):  # Read until EOF
            callback(line)


def run_command(command: list[str]) -> tuple[int, str, str]:
    """Run a subprocess command, capturing its stdout, stderr, and return code."""
    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )

    stdout_lines = []
    stderr_lines = []

    def print_and_store_stdout(line):
        print(line, end="")  # Print in real-time
        stdout_lines.append(line)

    def print_and_store_stderr(line):
        print(line, end="")  # Print in real-time
        stderr_lines.append(line)

    # Create threads to read stdout and stderr
    stdout_thread = threading.Thread(
        target=_stream_reader,
        args=(process.stdout, print_and_store_stdout),
        daemon=True,
    )
    stderr_thread = threading.Thread(
        target=_stream_reader,
        args=(process.stderr, print_and_store_stderr),
        daemon=True,
    )

    stdout_thread.start()
    stderr_thread.start()

    return_code = process.wait()  # Ensure process completes

    stdout_thread.join()
    stderr_thread.join()

    return return_code, "".join(stdout_lines), "".join(stderr_lines)


@retry(10, (LockError,), delay=0.5)
def repro(name: str, force: bool) -> tuple[int, str, str]:
    """Reproduce a DVC stage.

    Parameters
    ----------
        name : str
            The name of the stage to reproduce.

    Returns
    -------
        Tuple[int, str, str]
            The return code, stdout, and stderr of the process.
    """
    stdout_lines = []
    stderr_lines = []

    # Run the main repro command
    cmd = ["dvc", "repro", "--single-item", name]
    if force:
        cmd.append("--force")
    return_code, repro_stdout, repro_stderr = run_command(cmd)
    stdout_lines.append(repro_stdout)
    stderr_lines.append(repro_stderr)

    # Handle lock errors in repro
    if "ERROR: Unable to acquire lock" in repro_stderr:
        raise LockError(f"Unable to acquire lock for {name}")
    if f"ERROR: failed to reproduce '{name}': Unable to acquire lock" in repro_stderr:
        for _ in range(5):
            try:
                print(f"Committing {name} again due to lock error")
                commit_code, commit_stdout, commit_stderr = run_command(
                    ["dvc", "commit", name, "--force"]
                )
                stdout_lines.append(commit_stdout)
                stderr_lines.append(commit_stderr)
                if commit_code == 0:
                    return_code = 0  # we were able to commit the lock
                    break
            except subprocess.CalledProcessError:
                time.sleep(0.5)
        else:
            raise LockError(f"Unable to commit lock for {name}")

    # Combine and return outputs
    return return_code, "".join(stdout_lines), "".join(stderr_lines)


@retry(10, (LockError,), delay=0.5)
def checkout(
    stage_lock: dict, cached_job_lock_json: str, name: str
) -> tuple[int, str, str]:
    log.info(f"Checking out job '{name}'")
    cached_job_lock = json.loads(cached_job_lock_json)
    output_lock = transform_lock(stage_lock, cached_job_lock)

    stdout_lines = []
    stderr_lines = []

    lock_file = Path("dvc.lock")
    if not lock_file.exists():
        with lock_file.open("w") as f:
            yaml.dump(
                {
                    "schema": "2.0",
                    "stages": {},
                },
                f,
            )

    fs = dvc.api.DVCFileSystem(url=None, rev=None)
    with fs.repo.lock:  # this can raise a LockError directly
        with lock_file.open("r") as f:
            lock = yaml.safe_load(f)
            lock["stages"][name] = output_lock

        with lock_file.open("w") as f:
            stdout_lines.append(f"Updating lock file 'dvc.lock' for '{name}'\n")
            yaml.dump(lock, f)

    # Run the main repro command
    # We can use force here, because `dvc repro` would also remove the files
    stdout_lines.append(f"Checking out stage '{name}':\n")
    return_code, repro_stdout, repro_stderr = run_command(
        ["dvc", "checkout", "--force", name]
    )

    if "ERROR: Unable to acquire lock" in repro_stderr:
        # here we raise a lock error, because the subprocess was
        # unable to acquire the lock
        raise LockError(f"Unable to acquire lock for checking out {name}.")

    # dvc checkout does not raise any error for GIT tracked files
    # therefore, we run ``dvc status`` to check if the checkout was successful
    if return_code == 0:
        return_code, status_stdout, status_stderr = run_command(["dvc", "status", name])
        stdout_lines.append(status_stdout)
        stderr_lines.append(status_stderr)
        if "Data and pipelines are up to date." not in status_stdout:
            # TODO: we need to run `dvc repro` instead of checkout.
            return 404, "".join(stdout_lines), "".join(stderr_lines)

    stdout_lines.append(repro_stdout)
    stderr_lines.append(repro_stderr)

    return return_code, "".join(stdout_lines), "".join(stderr_lines)



import datetime
import fnmatch
import json
from typing import List, Literal, Optional

import networkx as nx
from dvc.stage.cache import _get_cache_hash
from sqlmodel import (
    Field,
    Relationship,
    Session,
    SQLModel,
    String,
    create_engine,
    or_,
    select,
)

from paraffin.lock import clean_lock
from paraffin.stage import PipelineStageDC
from paraffin.utils import get_group


class Worker(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str
    machine: str
    status: Literal["running", "idle", "offline"] = Field(
        sa_type=String, default="idle"
    )
    last_seen: datetime.datetime = Field(default_factory=datetime.datetime.now)

    # Relationships
    jobs: List["Job"] = Relationship(back_populates="worker")
    cwd: str = ""  # Current working directory
    pid: int = 0  # Process ID
    started_at: datetime.datetime = Field(default_factory=datetime.datetime.now)
    finished_at: Optional[datetime.datetime] = None


class JobDependency(SQLModel, table=True):
    parent_id: Optional[int] = Field(foreign_key="job.id", primary_key=True)
    child_id: Optional[int] = Field(foreign_key="job.id", primary_key=True)


class Experiment(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    base: str  # Commit this experiment is based on
    origin: str = "local"  # Origin of the repository, e.g. https://...
    machine: str = "local"  # Machine where the experiment was submitted from
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)

    # Relationships
    jobs: List["Job"] = Relationship(back_populates="experiment")


class Job(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str
    cmd: str  # Command to execute
    status: Literal["pending", "running", "completed", "cached", "failed"] = Field(
        sa_type=String, default="pending"
    )
    queue: str = "default"  # Queue name(s)
    lock: str = ""  # JSON string of lockfile for the stage
    deps_hash: str = ""  # Hash of the dependencies
    experiment_id: Optional[int] = Field(foreign_key="experiment.id")
    stderr: str = ""  # stderr output
    stdout: str = ""  # stdout output
    capture_stderr: bool = True  # Capture stderr output
    capture_stdout: bool = True  # Capture stdout output
    started_at: Optional[datetime.datetime] = None
    finished_at: Optional[datetime.datetime] = None
    worker_id: Optional[int] = Field(foreign_key="worker.id", default=None)
    cache: bool = False  # Use the paraffin cache for this job
    force: bool = False  # rerun the job even if cached

    # Relationships
    experiment: Optional[Experiment] = Relationship(back_populates="jobs")
    worker: Optional[Worker] = Relationship(back_populates="jobs")
    parents: List["Job"] = Relationship(
        link_model=JobDependency,
        back_populates="children",
        sa_relationship_kwargs={
            "primaryjoin": "Job.id==JobDependency.child_id",
            "secondaryjoin": "Job.id==JobDependency.parent_id",
        },
    )
    children: List["Job"] = Relationship(
        link_model=JobDependency,
        back_populates="parents",
        sa_relationship_kwargs={
            "primaryjoin": "Job.id==JobDependency.parent_id",
            "secondaryjoin": "Job.id==JobDependency.child_id",
        },
    )


def save_graph_to_db(
    graph: nx.DiGraph,
    queues: dict[str, str],
    commit: str,
    origin: str,
    machine: str,
    cache: bool,
    db_url: str,
) -> None:
    engine = create_engine(db_url)
    SQLModel.metadata.create_all(engine)
    with Session(engine) as session:
        experiment = Experiment(base=commit, origin=origin, machine=machine)
        session.add(experiment)
        session.commit()
        for node in nx.topological_sort(graph):
            node: PipelineStageDC
            queue = "default"
            # use fnmatch to match the node name with the custom queues
            for pattern, q in queues.items():
                if fnmatch.fnmatch(node.name, pattern):
                    queue = q
                    break
            status = "pending" if node.changed else "cached"

            job = Job(
                cmd=json.dumps(node.cmd),
                name=node.name,
                queue=queue,
                status=status,
                experiment_id=experiment.id,
                cache=cache,
                force=node.force,
            )
            # if completed, we can look for the lock and deps_hash
            if status == "completed":
                # TODO: get the lock and deps_hash from the stage
                pass

            session.add(job)
            # add dependencies
            for parent in graph.predecessors(node):
                parent_job = session.exec(
                    select(Job)
                    .where(Job.experiment_id == experiment.id)
                    .where(Job.name == parent.name)
                ).one()
                session.add(JobDependency(parent_id=parent_job.id, child_id=job.id))

        session.commit()


def list_experiments(db_url: str, commit: str | None) -> list[dict]:
    engine = create_engine(db_url)
    with Session(engine) as session:
        if commit is not None:
            exps = session.exec(
                select(Experiment).where(Experiment.base == commit)
            ).all()
        else:
            exps = session.exec(select(Experiment)).all()
        return [exp.model_dump() for exp in exps]


def session_to_graph(session: Session, experiment_id: int | None) -> nx.DiGraph:
    """
    Create a directed graph from jobs in the database session,
      keeping Job objects as node data.
    """
    statement = select(Job)
    if experiment_id:
        statement = statement.where(Job.experiment_id == experiment_id)
    jobs = session.exec(statement).all()

    graph = nx.DiGraph()
    for job in jobs:
        graph.add_node(job.id, data=job)
        for parent in job.parents:
            graph.add_edge(parent.id, job.id)

    return graph


def db_to_graph(db_url: str, experiment_id: int = 1) -> nx.DiGraph:
    """
    Create a directed graph from the database for a specific experiment,
      resolving Job objects to dictionaries.
    """
    engine = create_engine(db_url)
    with Session(engine) as session:
        # Create the graph using the open session
        graph = session_to_graph(session, experiment_id)

        # Resolve Job objects to dictionaries
        resolved_graph = nx.DiGraph()
        for job_id, node_data in graph.nodes(data=True):
            job = node_data["data"]
            resolved_graph.add_node(
                job_id,
                name=job.name,
                cmd=json.loads(job.cmd),
                status=job.status,
                queue=job.queue,
                lock=json.loads(job.lock) if job.lock else None,
                deps_hash=job.deps_hash,
                group=get_group(job.name)[0],
            )

        # Add edges from the original graph
        resolved_graph.add_edges_from(graph.edges(data=True))

        return resolved_graph


def get_job(
    db_url: str,
    worker_id: int,
    queues: list | None = None,
    experiment: int | None = None,
    job_name: str | None = None,
) -> dict | None:
    """
    Get the next job where status is 'pending' and all parents are 'completed'.
    """
    engine = create_engine(db_url)
    with Session(engine) as session:
        if job_name is None:
            results = _fetch_pending_jobs(session, experiment, queues)
        else:
            results = _fetch_jobs_by_name(session, experiment, queues, job_name)

        for job in results:
            if _all_parents_completed(job):
                _update_job_status(session, job, worker_id)
                return _job_to_dict(job)
    return None


def _fetch_pending_jobs(
    session: Session, experiment: int | None, queues: list | None
) -> list:
    """
    Fetch jobs with 'pending' or 'cached' status, optionally
    filtered by experiment and queues.
    """
    statement = select(Job).where(or_(Job.status == "pending", Job.status == "cached"))
    if experiment:
        statement = statement.where(Job.experiment_id == experiment)
    if queues:
        statement = statement.where(Job.queue.in_(queues))
    statement = statement.with_for_update()
    return session.exec(statement).all()


def _fetch_jobs_by_name(
    session: Session, experiment: int | None, queues: list | None, job_name: str
) -> list:
    """
    Fetch jobs by name, including their predecessors, and filter by status and queues.
    """
    graph = session_to_graph(session, experiment)
    statement = select(Job).where(Job.name == job_name)
    if experiment:
        statement = statement.where(Job.experiment_id == experiment)
    jobs = session.exec(statement).all()

    results = []
    for job in jobs:
        predecessors = nx.ancestors(graph, job.id)
        results.extend(
            [graph.nodes[node]["data"] for node in graph.nodes if node in predecessors]
        )
        results.append(job)

    # Filter results that are not "pending" or "cached"
    results = [job for job in results if job.status in ["pending", "cached"]]
    if queues:
        results = [job for job in results if job.queue in queues]
    return results


def _all_parents_completed(job) -> bool:
    """
    Check if all parents of a job are completed.
    """
    return all(parent.status == "completed" for parent in job.parents)


def _update_job_status(session: Session, job, worker_id: int) -> None:
    """
    Update the job status to 'running' and set the worker ID and start time.
    """
    job.status = "running"
    job.started_at = datetime.datetime.now()
    job.worker_id = worker_id
    session.add(job)
    session.commit()


def _job_to_dict(job) -> dict:
    """
    Convert a Job object to a dictionary.
    """
    return {
        "id": job.id,
        "name": job.name,
        "cmd": json.loads(job.cmd),
        "queue": job.queue,
        "status": job.status,
        "cache": job.cache,
        "force": job.force,
    }


def complete_job(
    job_id: int,
    lock: dict,
    db_url: str,
    status: str = "completed",
    stderr: str = "",
    stdout: str = "",
):
    engine = create_engine(db_url)
    with Session(engine) as session:
        statement = select(Job).where(Job.id == job_id)
        results = session.exec(statement)
        job = results.one()
        job.status = status
        job.lock = json.dumps(lock)
        if job.capture_stderr:
            job.stderr = stderr
        if job.capture_stdout:
            job.stdout = stdout
        job.finished_at = datetime.datetime.now()
        # We only write the deps_hash to the database
        #  once the job has finished successfully!
        if status == "completed":
            job.deps_hash = _get_cache_hash(clean_lock(lock), key=False)
        session.add(job)
        session.commit()


def update_job_status(
    job_name: str, experiment_id: int, status: str, db_url: str, force: bool
) -> int:
    engine = create_engine(db_url)
    with Session(engine) as session:
        statement = (
            select(Job)
            .where(Job.experiment_id == experiment_id)
            .where(Job.name == job_name)
        )
        results = session.exec(statement)
        job = results.one()
        if job.status == "completed" and not force:
            return -1
        job.status = status
        if force:
            job.force = True
        session.add(job)
        session.commit()
    return 0


def get_job_dump(job_name: str, experiment_id: int, db_url: str) -> dict[str, str]:
    engine = create_engine(db_url)
    with Session(engine) as session:
        statement = (
            select(Job)
            .where(Job.experiment_id == experiment_id)
            .where(Job.name == job_name)
        )
        results = session.exec(statement)
        job = results.one()
        data = job.model_dump()
        if job.worker:
            data.update({"worker": job.worker.model_dump()})
        return data


def find_cached_job(db_url: str, deps_cache: str = "") -> dict:
    engine = create_engine(db_url)
    with Session(engine) as session:
        statement = select(Job).where(Job.deps_hash == deps_cache)
        results = session.exec(statement)
        if res := results.first():
            return res.model_dump()
    return {}


def register_worker(name: str, machine: str, db_url: str, cwd: str, pid: int) -> int:
    engine = create_engine(db_url)
    with Session(engine) as session:
        worker = Worker(name=name, machine=machine, cwd=cwd, pid=pid)
        session.add(worker)
        session.commit()
        return worker.id


def update_worker(id: int, status: str, db_url: str) -> None:
    engine = create_engine(db_url)
    with Session(engine) as session:
        worker = session.exec(select(Worker).where(Worker.id == id)).one()
        worker.status = status
        worker.last_seen = datetime.datetime.now()
        session.add(worker)
        session.commit()


def close_worker(id: int, db_url: str) -> None:
    engine = create_engine(db_url)
    with Session(engine) as session:
        worker = session.exec(select(Worker).where(Worker.id == id)).one()
        worker.status = "offline"
        worker.last_seen = datetime.datetime.now()
        worker.finished_at = datetime.datetime.now()
        session.add(worker)
        session.commit()


def list_workers(db_url: str, id: int | None = None) -> list[dict]:
    engine = create_engine(db_url)
    with Session(engine) as session:
        if id is None:
            statement = select(Worker).where(Worker.status != "offline")
            workers = session.exec(statement).all()
        else:
            workers = session.exec(select(Worker).where(Worker.id == id)).all()
        data = []
        for worker in workers:
            _data = worker.model_dump()
            _data["jobs"] = [job.id for job in worker.jobs]
            data.append(_data)
        return data


def get_jobs(db_url: str, experiment_id: int) -> dict[str, int]:
    """Get the number of jobs in each status for a specific experiment."""
    engine = create_engine(db_url)
    with Session(engine) as session:
        statement = select(Job).where(Job.experiment_id == experiment_id)
        jobs = session.exec(statement).all()

        status = {"pending": 0, "running": 0, "completed": 0, "cached": 0, "failed": 0}
        for job in jobs:
            status[job.status] += 1

        return status



import pytest
from dvc.stage.cache import _get_cache_hash

from paraffin.lock import clean_lock, transform_lock


@pytest.fixture()
def lock01() -> tuple[dict, dict]:
    raw = {
        "cmd": "zntrack run zntrack.examples.AddNodeNumbers --name A_AddNodeNumbers",
        "params": {"params.yaml": {"A_AddNodeNumbers": {"params": 644999}}},
        "deps": [
            {
                "path": "nodes/A/1/ParamsToOuts/outs.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
            {
                "path": "nodes/A/1/ParamsToOuts_1/outs.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
        "outs": [
            {
                "path": "nodes/A/AddNodeNumbers/node-meta.json",
                "hash": "md5",
                "md5": "461013724d26fce139a6586d6635b42c",
            },
            {
                "path": "nodes/A/AddNodeNumbers/sum.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
    }

    exp = {
        "cmd": "zntrack run zntrack.examples.AddNodeNumbers --name <node-name>",
        "params": {"params.yaml": {"<node-name>": {"params": 644999}}},
        "deps": [
            {
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
            {
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
    }

    return raw, exp


@pytest.fixture()
def lock02() -> tuple[dict, dict]:
    raw = {
        "cmd": "zntrack run package.MyNode --name MyNode",
        "params": {
            "cp2k.yaml": {"basis": "DZVP-MOLOPT-SR-GTH"},
            "params.yaml": {"MyNode": {"params": 644999}},
        },
        "outs": [
            {
                "path": "nodes/MyNode/node-meta.json",
                "hash": "md5",
                "md5": "461013724d0429e139a6586d6635b42c",
            },
        ],
    }

    exp = {
        "cmd": "zntrack run package.MyNode --name <node-name>",
        "params": {
            "cp2k.yaml": {"basis": "DZVP-MOLOPT-SR-GTH"},
            "params.yaml": {"<node-name>": {"params": 644999}},
        },
    }

    return raw, exp


@pytest.fixture()
def lock03() -> tuple[dict, dict]:
    raw = {
        "cmd": "zntrack run package.MyNode --name MyNode",
        "outs": [
            {
                "path": "nodes/MyNode/node-meta.json",
                "hash": "md5",
                "md5": "461013724d0429e139a6586d6635b42c",
            },
        ],
    }

    exp = {
        "cmd": "zntrack run package.MyNode --name <node-name>",
    }

    return raw, exp


@pytest.fixture()
def lock04() -> tuple[dict, dict]:
    raw = {
        "cmd": ["zntrack run package.MyNode --name MyNode"],
        "outs": [
            {
                "path": "nodes/MyNode/node-meta.json",
                "hash": "md5",
                "md5": "461013724d0429e139a6586d6635b42c",
            },
        ],
    }

    exp = {
        "cmd": ["zntrack run package.MyNode --name <node-name>"],
    }

    return raw, exp


@pytest.fixture()
def lock05() -> tuple[dict, dict]:
    raw = {
        "cmd": "zntrack run package.MyNode --name MyNode",
        "deps": [
            {
                "path": "data",
                "hash": "md5",
                "files": [
                    {
                        "relpath": "data.csv",
                        "md5": "421109828f8547af8727ca039ebd3d13",
                        "size": 23,
                    }
                ],
            },
            {
                "path": "nodes/MyNode/outs.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
    }

    exp = {
        "cmd": "zntrack run package.MyNode --name <node-name>",
        "deps": [
            {
                "hash": "md5",
                "files": [
                    {
                        "md5": "421109828f8547af8727ca039ebd3d13",
                        "relpath": "data.csv",
                        # relpath should be fine, because it won't contain the node name
                        "size": 23,
                    }
                ],
            },
            {
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
    }

    return raw, exp


@pytest.fixture()
def lock_a_b() -> tuple[dict, dict]:
    """Two locks that should yield the same hash"""

    a = {
        "cmd": "zntrack run package.MyNode --name MyNode_1",
        "params": {"params.yaml": {"MyNode_1": {"params": 644999}}},
        "deps": [
            {
                "path": "nodes/SomeNodeA/ParamsToOuts_1/outs.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
        "outs": [
            {
                "path": "nodes/MyNode_2/node-meta.json",
                "hash": "md5",
                "md5": "461013724d0429e139a6586d6635b42c",
            },
        ],
    }

    b = {
        "cmd": "zntrack run package.MyNode --name MyNode_2",
        "params": {"params.yaml": {"MyNode_2": {"params": 644999}}},
        "deps": [
            {
                "path": "nodes/SomeNodeB/ParamsToOuts_2/outs.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
        "outs": [
            {
                "path": "nodes/MyNode_1/node-meta.json",
                "hash": "md5",
                "md5": "461013724d0429e139a6586d6635b42c",
            },
        ],
    }

    return a, b


@pytest.fixture()
def lock_input_ref_output() -> tuple[dict, dict, dict]:
    """
    Returns
    -------
    tuple[dict, dict, dict]
        Input hash only contains the lock of a not-yet executed stage but
        with the deps and params.
        Ref hash contains the lock of a stage that has been executed and
        is being loaded from the database.
        Output hash is the expected new lock from the input that should
        be written to "dvc.lock" to load the correct data.
    """
    inp = {
        "cmd": "zntrack run package.MyNode --name grp_MyNode",
        "params": {"params.yaml": {"grp_MyNode": {"params": 644999}}},
        "deps": [
            {
                "path": "nodes/ParamsToOuts/outs.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
    }

    ref = {
        "cmd": "zntrack run package.MyNode --name MyNode_1",
        "params": {"params.yaml": {"MyNode_1": {"params": 644999}}},
        "deps": [
            {
                "path": "nodes/SomeNodeA/ParamsToOuts_1/outs.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
        "outs": [
            {
                "path": "nodes/MyNode_1/node-meta.json",
                "hash": "md5",
                "md5": "461013724d26fce139a6586d6635b42c",
            },
            {
                "path": "nodes/MyNode_1/sum.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
    }

    out = {
        "cmd": "zntrack run package.MyNode --name grp_MyNode",
        "params": {"params.yaml": {"grp_MyNode": {"params": 644999}}},
        "deps": [
            {
                "path": "nodes/ParamsToOuts/outs.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
        "outs": [
            {
                "path": "nodes/grp/MyNode/node-meta.json",
                "hash": "md5",
                "md5": "461013724d26fce139a6586d6635b42c",
            },
            {
                "path": "nodes/grp/MyNode/sum.json",
                "hash": "md5",
                "md5": "fb6d880180fbf208fab297f75d32c5ce",
            },
        ],
    }

    return inp, ref, out


def test_cache_hash(lock01):
    assert (
        "ad8ecfc775a1d13b602acbfa3cf563f6616ca9d76ee25e8cf45fac390742743e"
        == _get_cache_hash(lock01[0], key=False)
    )
    assert (
        "10894a16d56e5f45557a481820ce0e188d7570ce900c31c3ddbbb0d325901bf5"
        == _get_cache_hash(lock01[1], key=False)
    )


def test_clean_lock01(lock01):
    raw, expected = lock01
    assert clean_lock(raw) == expected


def test_clean_lock02(lock02):
    raw, expected = lock02
    assert clean_lock(raw) == expected


def test_clean_lock03(lock03):
    raw, expected = lock03
    assert clean_lock(raw) == expected


def test_clean_lock04(lock04):
    raw, expected = lock04
    assert clean_lock(raw) == expected


def test_clean_lock05(lock05):
    raw, expected = lock05
    assert clean_lock(raw) == expected


def test_clean_lockAB(lock_a_b):
    a, b = lock_a_b
    clean_a = clean_lock(a)
    clean_b = clean_lock(b)

    assert clean_a == clean_b
    assert _get_cache_hash(clean_a, key=False) == _get_cache_hash(clean_b, key=False)


def test_transform_lock01(lock_input_ref_output):
    inp, ref, out = lock_input_ref_output
    assert transform_lock(inp, ref) == out



import os
import subprocess

import git
import pytest
import yaml
from typer.testing import CliRunner

from paraffin.cli import app

runner = CliRunner()


DVC_YAML = {
    "stages": {
        "00a_syncdata": {
            "outs": ["output"],
            "cmd": [
                """bash -euo pipefail << EOF

bash src/00_syncdata.sh

EOF"""
            ],
        }
    }
}


@pytest.fixture
def repo(tmp_path):
    os.chdir(tmp_path)
    repo = git.Repo.init(tmp_path)
    # run dvc init in the repo
    subprocess.run(["dvc", "init"], cwd=tmp_path)
    (tmp_path / "dvc.yaml").write_text(yaml.dump(DVC_YAML))
    (tmp_path / "src").mkdir()
    (tmp_path / "src" / "00_syncdata.sh").write_text(
        """#!/bin/bash
echo "Hello, World!" > output
"""
    )
    repo.index.add(["dvc.yaml", "src/00_syncdata.sh"])
    repo.index.commit("Initial commit")
    return repo


def test_051(repo):
    # Test an issue where `cmd` is a list instead of a string
    subprocess.check_call(["dvc repro"], shell=True)
    subprocess.check_call(["dvc freeze 00a_syncdata"], shell=True)

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0

    result = runner.invoke(app, "worker")
    assert result.exit_code == 0



import random

import zntrack
from typer.testing import CliRunner

from paraffin.cli import app

runner = CliRunner()


class A(zntrack.Node):
    metrics: dict = zntrack.metrics()

    def run(self):
        self.metrics = {"random": random.random()}


class B(zntrack.Node):
    a: A = zntrack.deps()
    metrics: dict = zntrack.metrics()

    def run(self):
        self.metrics = {"random": self.a.metrics["random"]}


def test_single_item(proj_path, check_finished):
    project = zntrack.Project()

    with project:
        a = A()
        B(a=a)

    project.build()

    result = runner.invoke(app, "submit --single-item B")
    assert result.exit_code == 0
    result = runner.invoke(app, "worker")
    # single item submit should fail, as B depends on A which is missing
    assert "ERROR: failed to reproduce 'B'" in result.stdout

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    result = runner.invoke(app, "worker")
    assert result.exit_code == 0

    assert check_finished()

    assert A.from_rev().metrics["random"] == B.from_rev().metrics["random"]
